# -*- coding: utf-8 -*-

# Copyright © 2017-2019 rfgb Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program (at the base of this repository). If not,
# see <http://www.gnu.org/licenses/>

"""
_metadata.py
============

Metadata for rfgb package
"""

__author__ = "Alexander L. Hayes"
__maintainer__ = "Alexander L. Hayes"
__email__ = "hayesall@iu.edu"
__copyright__ = "Copyright © 2017-2019 rfgb Contributors"
__license__ = "GNU General Public License v3.0 (GPLv3)"
__version__ = "0.3.0"
__status__ = "Prototype"
